@@ -40,7 +40,7 @@
 #include <mipict.h>
 
 #if 0
-#define __DBG(x) ErrorF x
+#define __DBG(x) LogF x
 #else
 #define __DBG(x)
 #endif
@@ -119,9 +119,9 @@ static void _assert_pixmap_contains_box(PixmapPtr pixmap, BoxPtr box, const char
 
 static void apply_damage(struct sna_composite_op *op, RegionPtr region)
 {
-	DBG(("%s: damage=%p, region=%ldx[(%d, %d), (%d, %d)]\n",
+	DBG(("%s: damage=%p, region=%dx[(%d, %d), (%d, %d)]\n",
 	     __FUNCTION__, op->damage,
-	     (long)REGION_NUM_RECTS(region),
+	     region_num_rects(region),
 	     region->extents.x1, region->extents.y1,
 	     region->extents.x2, region->extents.y2));
 
@@ -390,17 +390,17 @@ mono_merge_unsorted_edges(struct mono_edge *head, struct mono_edge *unsorted)
 static inline void
 __dbg_mono_edges(const char *function, struct mono_edge *edges)
 {
-	ErrorF("%s: ", function);
+	DBG(("%s: ", function));
 	while (edges) {
 		if (edges->x.quo < INT16_MAX << 16) {
-			ErrorF("(%d.%06d)+(%d.%06d)x%d, ",
-			       edges->x.quo, edges->x.rem,
-			       edges->dxdy.quo, edges->dxdy.rem,
-			       edges->dy*edges->dir);
+			DBG(("(%d.%06d)+(%d.%06d)x%d, ",
+			     edges->x.quo, edges->x.rem,
+			     edges->dxdy.quo, edges->dxdy.rem,
+			     edges->dy*edges->dir));
 		}
 		edges = edges->next;
 	}
-	ErrorF("\n");
+	DBG(("\n"));
 }
 #define DBG_MONO_EDGES(x) __dbg_mono_edges(__FUNCTION__, x)
 static inline void
@@ -444,10 +444,10 @@ mono_span(struct mono *c, int x1, int x2, BoxPtr box)
 
 		pixman_region_init_rects(&region, box, 1);
 		RegionIntersect(&region, &region, &c->clip);
-		if (REGION_NUM_RECTS(&region)) {
+		if (region_num_rects(&region)) {
 			c->op.boxes(c->sna, &c->op,
-				    REGION_RECTS(&region),
-				    REGION_NUM_RECTS(&region));
+				    region_rects(&region),
+				    region_num_rects(&region));
 			apply_damage(&c->op, &region);
 		}
 		pixman_region_fini(&region);
@@ -505,10 +505,10 @@ thread_mono_span_clipped(struct mono *c, int x1, int x2, BoxPtr box)
 
 	pixman_region_init_rects(&region, box, 1);
 	RegionIntersect(&region, &region, &c->clip);
-	if (REGION_NUM_RECTS(&region))
+	if (region_num_rects(&region))
 		thread_mono_span_add_boxes(c,
-					   REGION_RECTS(&region),
-					   REGION_NUM_RECTS(&region));
+					   region_rects(&region),
+					   region_num_rects(&region));
 	pixman_region_fini(&region);
 }
 
@@ -769,8 +769,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 
 	trapezoid_origin(&traps[0].left, &dst_x, &dst_y);
 
-	trapezoids_bounds(ntrap, traps, &extents);
-	if (extents.y1 >= extents.y2 || extents.x1 >= extents.x2)
+	if (!trapezoids_bounds(ntrap, traps, &extents))
 		return true;
 
 	DBG(("%s: extents (%d, %d), (%d, %d)\n",
@@ -803,6 +802,9 @@ mono_trapezoids_span_converter(struct sna *sna,
 	unbounded = (!sna_drawable_is_clear(dst->pDrawable) &&
 		     !operator_is_bounded(op));
 
+	if (op == PictOpClear && sna->clear)
+		src = sna->clear;
+
 	mono.sna = sna;
 	if (!mono.sna->render.composite(mono.sna, op, src, NULL, dst,
 				       src_x + mono.clip.extents.x1 - dst_x - dx,
@@ -811,7 +813,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 				       mono.clip.extents.x1,  mono.clip.extents.y1,
 				       mono.clip.extents.x2 - mono.clip.extents.x1,
 				       mono.clip.extents.y2 - mono.clip.extents.y1,
-				       memset(&mono.op, 0, sizeof(mono.op))))
+				       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op))))
 		return false;
 
 	num_threads = 1;
@@ -850,7 +852,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 			threads[n].extents.y1 = y;
 			threads[n].extents.y2 = y += h;
 
-			sna_threads_run(mono_span_thread, &threads[n]);
+			sna_threads_run(n, mono_span_thread, &threads[n]);
 		}
 
 		threads[0].extents.y1 = y;
@@ -922,7 +924,6 @@ mono_trapezoids_span_converter(struct sna *sna,
 				      traps[n].top, traps[n].bottom,
 				      &traps[n].right.p1, &traps[n].right.p2, -1);
 		}
-		memset(&mono.op, 0, sizeof(mono.op));
 		if (mono.sna->render.composite(mono.sna,
 					       PictOpClear,
 					       mono.sna->clear, NULL, dst,
@@ -931,7 +932,7 @@ mono_trapezoids_span_converter(struct sna *sna,
 					       mono.clip.extents.x1,  mono.clip.extents.y1,
 					       mono.clip.extents.x2 - mono.clip.extents.x1,
 					       mono.clip.extents.y2 - mono.clip.extents.y1,
-					       &mono.op)) {
+					       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 			mono_render(&mono);
 			mono.op.done(mono.sna, &mono.op);
 		}
@@ -1046,9 +1047,7 @@ mono_trapezoid_span_inplace(struct sna *sna,
 	int was_clear;
 	int x, y, n;
 
-	trapezoids_bounds(ntrap, traps, &mono.clip.extents);
-	if (mono.clip.extents.y1 >= mono.clip.extents.y2 ||
-	    mono.clip.extents.x1 >= mono.clip.extents.x2)
+	if (!trapezoids_bounds(ntrap, traps, &mono.clip.extents))
 		return true;
 
 	DBG(("%s: extents (%d, %d), (%d, %d)\n",
@@ -1169,7 +1168,10 @@ unbounded_pass:
 		mono.span = mono_span__fast;
 	else
 		mono.span = mono_span;
-	mono_render(&mono);
+	if (sigtrap_get() == 0) {
+		mono_render(&mono);
+		sigtrap_put();
+	}
 	mono_fini(&mono);
 
 	if (op) {
@@ -1273,14 +1275,13 @@ mono_trap_span_converter(struct sna *sna,
 			      &p1, &p2, -1);
 	}
 
-	memset(&mono.op, 0, sizeof(mono.op));
 	if (mono.sna->render.composite(mono.sna, PictOpAdd, src, NULL, dst,
-					0, 0,
-					0, 0,
-					mono.clip.extents.x1,  mono.clip.extents.y1,
-					mono.clip.extents.x2 - mono.clip.extents.x1,
-					mono.clip.extents.y2 - mono.clip.extents.y1,
-					&mono.op)) {
+				       0, 0,
+				       0, 0,
+				       mono.clip.extents.x1,  mono.clip.extents.y1,
+				       mono.clip.extents.x2 - mono.clip.extents.x1,
+				       mono.clip.extents.y2 - mono.clip.extents.y1,
+				       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 		mono_render(&mono);
 		mono.op.done(mono.sna, &mono.op);
 	}
@@ -1356,7 +1357,6 @@ mono_triangles_span_converter(struct sna *sna,
 			      &tri[n].p3, &tri[n].p1, 1);
 	}
 
-	memset(&mono.op, 0, sizeof(mono.op));
 	if (mono.sna->render.composite(mono.sna, op, src, NULL, dst,
 				       src_x + mono.clip.extents.x1 - dst_x - dx,
 				       src_y + mono.clip.extents.y1 - dst_y - dy,
@@ -1364,7 +1364,7 @@ mono_triangles_span_converter(struct sna *sna,
 				       mono.clip.extents.x1,  mono.clip.extents.y1,
 				       mono.clip.extents.x2 - mono.clip.extents.x1,
 				       mono.clip.extents.y2 - mono.clip.extents.y1,
-				       &mono.op)) {
+				       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 		if (mono.clip.data == NULL && mono.op.damage == NULL)
 			mono.span = mono_span__fast;
 		else
@@ -1402,7 +1402,6 @@ mono_triangles_span_converter(struct sna *sna,
 				      &tri[n].p3, &tri[n].p1, 1);
 		}
 
-		memset(&mono.op, 0, sizeof(mono.op));
 		if (mono.sna->render.composite(mono.sna,
 					       PictOpClear,
 					       mono.sna->clear, NULL, dst,
@@ -1411,7 +1410,7 @@ mono_triangles_span_converter(struct sna *sna,
 					       mono.clip.extents.x1,  mono.clip.extents.y1,
 					       mono.clip.extents.x2 - mono.clip.extents.x1,
 					       mono.clip.extents.y2 - mono.clip.extents.y1,
-					       &mono.op)) {
+					       COMPOSITE_PARTIAL, memset(&mono.op, 0, sizeof(mono.op)))) {
 			if (mono.clip.data == NULL && mono.op.damage == NULL)
 				mono.span = mono_span__fast;
 			else
